#include "ntsc_video.h"
#include <string.h>

// The SPI channel
volatile SpiChannel spiChn = SPI_CHANNEL1 ;	// the SPI channel to use
volatile int spiClkDiv = 6 ;//5 MHz pixel rate => 256 points in 51.2 microSec

// main screen buffer array 256 wide x 200 high = 51200 pixels
// 51200/32 = 1600 integer words
int screen_buffer[1600];
volatile int *screen_buffer_addr = screen_buffer ;
volatile int *screen_ptr ;

// Current line number which is modified
// by a state machine in the timer2 ISR
volatile int LineCount = 0 ;
// ISR driven 1/60 second time
volatile int time_tick_60_hz = 0 ;

// video formatiing variables
//One bit masks for each bit of an INT
unsigned int pos[32] = {0x80000000,0x40000000,0x20000000,0x10000000,0x08000000,0x04000000,0x02000000,0x01000000,0x800000,0x400000,0x200000,0x100000,0x080000,0x040000,0x020000,0x010000,0x8000,0x4000,0x2000,0x1000,0x0800,0x0400,0x0200,0x0100,0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01};

// == OC3 ISR ============================================
// VECTOR 14 is OC3 vector -- set up of ipl 3 in main
// vector names from int_1xx_2xx.h
void __ISR(14, ipl3) OC3Handler(void) // 14
{
   // mPORTBSetBits(BIT_1);
    // Convert DMA to SPI control 
    DmaChnSetEventControl(DMAchn1, DMA_EV_START_IRQ(_SPI1_TX_IRQ)); //
    //DmaChnEnable(DMAchn1);
    // clear the timer interrupt flag -- name from
    // http://people.ece.cornell.edu/land/courses/ece4760/PIC32/Microchip_stuff/32-bit-Peripheral-Library-Guide.pdf
    // Table 8.2
    mOC3ClearIntFlag();
   // mPORTBClearBits(BIT_1);  // for profiling the ISR execution time
}

// == Timer 2 ISR =========================================
void __ISR(_TIMER_2_VECTOR, ipl2) Timer2Handler(void)
{
    //mPORTBSetBits(BIT_1); // for profiling the ISR execution time
    // update the current scanline number
    LineCount++ ;

    // start the DMA byte blaster to the screen
    if (LineCount >= image_start && LineCount < image_end){
        // set the Chan1 DMA transfer parameters: source & destination address,
        // source & destination size, number of bytes per event
        // 32 bytes / line with 4 bytes per transfer (SPI in 32 bit mode)
        //screen_ptr = screen_buffer + ((LineCount - image_start)<<5) ;
        DmaChnSetTxfer(DMAchn1, (void*)screen_ptr, (void*)&SPI1BUF, 32, 4, 4); //32
        // IRO 17 is the output compare 3 interrupt (See datasheet table 7.1)
        DmaChnSetEventControl(DMAchn1, DMA_EV_START_IRQ(17)); //
        // turn it on for 32 bytes
        DmaChnEnable(DMAchn1);
        // increment the image memory pointer for the next ISR pass
        screen_ptr += 8; // 8 32-bit words per line
    }

    // == SYNC state machine ====
    // begin long (Vertical) synch after line 247
    if (LineCount==248) {OC2R = line_cycles - us_5_cycles ;}
    // back to regular sync after line 250
    // the first condition eliminates sync for one line (to avoid duplicate)
    if (LineCount==250) {OC2R = 0 ;}
    if (LineCount==251) {OC2R = us_5_cycles ;}
    // start new frame after line 262 and reset the image memory pointer and
    // update the frame time_tick
    if (LineCount==263) {
        LineCount = 1;
        // reset for the next frame
        screen_ptr = screen_buffer_addr;
        // a general propose time base
        time_tick_60_hz++;
    }
    
    // clear the timer interrupt flag
    mT2ClearIntFlag();
    //mPORTBClearBits(BIT_1);  // for profiling the ISR execution time
}

void video_clr(){
    memset(screen_buffer, 0, sizeof(screen_buffer));
}

//== plot a point =========================================================
//plot one point
//at x,y with color 1=white 0=black 2=invert
#define word_per_line 8
void video_pt(int x, int y, char c) {
	//each line has 18 bytes
	//calculate i based upon this and x,y
	// the word with the pixel in it
	//int i = (x/32) + y*8
	int i = (x >> 5) + y * word_per_line ;
	if (x>=left && x<=right && y>=top && y<=bottom) {
		if (c==1)
	  	  screen_buffer[i] = screen_buffer[i] | pos[x & 0x1f];
        else if (c==0)
	  	  screen_buffer[i] = screen_buffer[i] & ~pos[x & 0x1f];
        else
	  	  screen_buffer[i] = screen_buffer[i] ^ pos[x & 0x1f];
	}
}

//==================================
//plot a line
//at x1,y1 to x2,y2 with color 1=white 0=black 2=invert
//NOTE: this function requires signed chars
//Code is from David Rodgers,
//"Procedural Elements of Computer Graphics",1985
void video_line(int x1, int y1, int x2, int y2, char c) {
	int e;
	signed int dx,dy,j, temp;
	signed char s1,s2, xchange;
        signed int x,y;

	x = x1;
	y = y1;

	//take absolute value
	if (x2 < x1) {
		dx = x1 - x2;
		s1 = -1;
	}

	else if (x2 == x1) {
		dx = 0;
		s1 = 0;
	}

	else {
		dx = x2 - x1;
		s1 = 1;
	}

	if (y2 < y1) {
		dy = y1 - y2;
		s2 = -1;
	}

	else if (y2 == y1) {
		dy = 0;
		s2 = 0;
	}

	else {
		dy = y2 - y1;
		s2 = 1;
	}

	xchange = 0;

	if (dy>dx) {
		temp = dx;
		dx = dy;
		dy = temp;
		xchange = 1;
	}

	e = ((int)dy<<1) - dx;

	for (j=0; j<=dx; j++) {
		video_pt(x,y,c);

		if (e>=0) {
			if (xchange==1) x = x + s1;
			else y = y + s2;
			e = e - ((int)dx<<1);
		}

		if (xchange==1) y = y + s2;
		else x = x + s1;

		e = e + ((int)dy<<1);
	}
}

//==================================
//return the value of one point
//at x,y with color 1=white 0=black 2=invert
char video_state(int x, int y) {
    //The following construction
    //detects exactly one bit at the x,y location
    // int i = (x>>3) + ((int)y<<4) + ((int)y<<3);
    int i = (x >> 5) + y * word_per_line ;
    return (screen_buffer[i] & pos[x & 0x1f]) != 0 ;
}

//==================================
// put a big character on the screen
// c is index into bitmap
void video_putchar(int x, int y, int c) {
    int i;
    int y_pos;
    int j;

    for (i=0;i<7;i++) {
        y_pos = y + i;

        j = ascii[c][i] ;

        video_pt(x,   y_pos, (j & 0x80)==0x80);
        video_pt(x+1, y_pos, (j & 0x40)==0x40);
        video_pt(x+2, y_pos, (j & 0x20)==0x20);
        video_pt(x+3, y_pos, (j & 0x10)==0x10);
        video_pt(x+4, y_pos, (j & 0x08)==0x08);
    }
}

//==================================
// put a string of big characters on the screen
void video_string(int x, int y, char *str) {
	char i;
	for (i=0; str[i]!=0; i++) {
		video_putchar(x,y,str[i]);
		x = x+6;
	}
}

//==================================
// put a big character on the screen
// c is index into bitmap
void video_big_putchar(int x, int y, int c) {
    int i;
    int y_pos;
    int j;

    for (i=0; i<7; i=i+1) {
        y_pos = y + i;

        j = ascii[c][i] ;

        video_pt(x,   y_pos, (j & 0x80)==0x80);
        video_pt(x+1, y_pos, (j & 0x80)==0x80);
        video_pt(x+2, y_pos, (j & 0x40)==0x40);
        video_pt(x+3, y_pos, (j & 0x40)==0x40);
        video_pt(x+4, y_pos, (j & 0x20)==0x20);
        video_pt(x+5, y_pos, (j & 0x40)==0x40);
        video_pt(x+6, y_pos, (j & 0x10)==0x10);
        video_pt(x+7, y_pos, (j & 0x10)==0x10);
        video_pt(x+8, y_pos, (j & 0x08)==0x08);
        video_pt(x+9, y_pos, (j & 0x08)==0x08);
    }
}

//==================================
// put a string of big characters on the screen
void video_bold_string(int x, int y, char *str) {
	char i;
	for (i=0; str[i]!=0; i++) {
		video_big_putchar(x,y,str[i]);
		x = x+10;
	}
}	

void init_ntsc_video(){
	// timer interrupt //////////////////////////
	// Set up timer2 on,  interrupts, internal clock, prescalar 1, toggle rate
	// at 40 MHz PB clock 63.5 microSec is
	// 2540 ticks
	OpenTimer2(T2_ON | T2_SOURCE_INT | T2_PS_1_1, line_cycles);
	// set up the timer interrupt with a priority of 2
	ConfigIntTimer2(T2_INT_ON | T2_INT_PRIOR_2);
	mT2ClearIntFlag(); // and clear the interrupt flag

	// Compuare match setup //////////////////////
	//Set up compare match unit to produce sync pulses
	// 5 uSec low
	// or 63.5-5 = 58.5 microSec (2340 ticks) low
	// pulse duration will be controlled in Timer2 ISR
	// #define OpenOC2( config, value1, value2) ( OC2RS = (value1), OC2R = (value2), OC2CON = (config) )
	OpenOC2(OC_ON | OC_TIMER2_SRC | OC_CONTINUE_PULSE, line_cycles-1, us_5_cycles);
	// OC2 is PPS group 2, map to RPB5 (pin 14)
	PPSOutput(2, RPB8, OC2);

	// OC3 setup /////////////////////////////////
	// Compare unit for video timing, 
	// using the interrupt flag to trigger the first DMA,
	// then use the ISR to change the DMA control to SPI
	// #define OpenOC2( config, value1, value2) ( OC2RS = (value1), OC2R = (value2), OC2CON = (config) )
	// Pulse needs to be TWO cycles long
	OpenOC3(OC_ON | OC_TIMER2_SRC | OC_CONTINUE_PULSE , us_10_cycles+2, us_10_cycles); //
	// turn on ISR so that DMA can covert to SPI control
	ConfigIntOC3(OC_INT_PRIOR_1 | OC_INT_ON); //3 //  
	mOC3ClearIntFlag(); // and clear the interrupt flag

	// SPI configure /////////////////////////////
	// SCK1 is pin 25 RB14
	// SDO1 is PPS group 2, map to RPA1 (pin 3)
	// SDI1 is PPS group 2, map to RPB8 (pin 17)
	// SS1 input is PPS group 1, map to RPB7 (pin 16) for framing
	// specify PPS group, signal, logical pin name
	//PPSInput (1, SS1, RPB7);
	PPSOutput(2, RPA1, SDO1);
	// control sync for DAC
	//mPORTBSetPinsDigitalOut(BIT_0 | BIT_1 | BIT_5);
	//mPORTBSetBits(BIT_0);

	// divide Fpb by spiClkDiv, configure the I/O ports.
	// 32 bit transfer
	SpiChnOpen(spiChn, SPI_OPEN_ON | SPI_OPEN_MODE32 | SPI_OPEN_MSTEN , spiClkDiv ) ;

	//=== DMA Channel 1 ================================
	// Open DMA Chan1 and chain from channel zero
	DmaChnOpen(DMAchn1, DMApri0, DMA_OPEN_DEFAULT);

  	// setup system wide interrupts  ///
  	INTEnableSystemMultiVectoredInt();
	return;								
}

void test_video_buffer(){
    int x;
    for (x = 0; x < 1600; x++){
        screen_buffer[x] = 0xcccccccc;
	}
}

void video_block(int x, int y, int width, int height, int* arr, char opt){
	int pos_i = 0;
	int i = 0;
	int x_i = x, y_i = y;
	int cur = 0;
	// iteration
	while (y_i < y+height){
		cur = (arr)[i];
		pos_i = 0;
		while(pos_i < 32){
			if (cur&pos[pos_i]){
				// there is a point at given position
				// display the point when it's in the screen range	
				if ((opt == OPT_OVERWRITE) || (opt == OPT_OVERLAY)) {
					if (x_i <= right && y_i <= bottom) video_pt(x_i, y_i, 1);
				} else if (opt == OPT_NEGATE) {
					if (x_i <= right && y_i <= bottom) video_pt(x_i, y_i, 2);
				} else {
                    if (x_i <= right && y_i <= bottom) video_pt(x_i, y_i, 0);
                }
			} else {
				// there is no point at given position
				// update based on option
				if (opt == OPT_OVERWRITE){
					video_pt(x_i, y_i, 0);
				}
			}
			// advance point by one
			// wrap around
			x_i ++;
			if (x_i >= x+width) {
				x_i = x;
				y_i ++;
			}
            if (y_i >= y+height) break;
			pos_i++;
		}
		i ++;
	}
}

void video_reverseblock(int x, int y, int width, int height, int* arr, char opt){
	int pos_i = 0;
	int i = 0;
	int x_i = x, y_i = y;
	int cur = 0;
	// iteration
	while (y_i < y+height){
		cur = (arr)[i];
		pos_i = 0;
		while(pos_i < 32){
			if (cur&pos[pos_i]){
				// there is a point at given position
				// display the point when it's in the screen range	
				if ((opt == OPT_OVERWRITE) || (opt == OPT_OVERLAY)) {
					video_pt((x*2-width-x_i), y_i, 1);
				} else if (opt == OPT_NEGATE) {
					video_pt((x*2-width-x_i), y_i, 2);
				} else {
                    video_pt((x*2-width-x_i), y_i, 0);
                }
			} else {
				// there is no point at given position
				// update based on option
				if (opt == OPT_OVERWRITE){
					video_pt(x_i, y_i, 0);
				}
			}
			// advance point by one
			// wrap around
			x_i ++;
			if (x_i >= x+width) {
				x_i = x;
				y_i ++;
			}
            if (y_i >= y+height) break;
			pos_i++;
		}
		i ++;
	}
}

void video_removechar(int x, int y, char c){
    int i;
    int y_pos;
    int j;

    for (i=0;i<7;i++) {
        y_pos = y + i;

        j = ascii[c][i] ;

        video_pt(x,   y_pos, ((j & 0x80)==0x80)^1);
        video_pt(x+1, y_pos, ((j & 0x40)==0x40)^1);
        video_pt(x+2, y_pos, ((j & 0x20)==0x20)^1);
        video_pt(x+3, y_pos, ((j & 0x10)==0x10)^1);
        video_pt(x+4, y_pos, ((j & 0x08)==0x08)^1);
    }
}
void video_removestring(int x, int y, char *str){
    char i;
	for (i=0; str[i]!=0; i++) {
		video_removechar(x,y,str[i]);
		x = x+6;
	}
}

void video_delchar(int x, int y, char c){
    int i;
    int y_pos;
    int j;

    for (i=0;i<7;i++) {
        y_pos = y + i;

        j = ascii[c][i] ;

        if ((j & 0x80)==0x80) video_pt(x,   y_pos, 2);
        if ((j & 0x40)==0x40) video_pt(x+1, y_pos, 2);
        if ((j & 0x20)==0x20) video_pt(x+2, y_pos, 2);
        if ((j & 0x10)==0x10) video_pt(x+3, y_pos, 2);
        if ((j & 0x08)==0x08) video_pt(x+4, y_pos, 2);
    }
}
void video_delstring(int x, int y, char *str){
    char i;
	for (i=0; str[i]!=0; i++) {
		video_delchar(x,y,str[i]);
		x = x+6;
	}
}


// drawing functions imported from tft_gfx library

// draw quick H and V lines
// opt: 0=black, 1=white, 2=invert
void video_drawFastHLine(short x0, short y0, short length, char opt){
	short x;
	for (x=x0; x<x0+length; x++){
		video_pt(x, y0, opt);
	}
}

void video_drawFastVLine(short x0, short y0, short length, char opt){
	short y;
	for (y=y0; y<y0+length; y++){
		video_pt(x0, y, opt);
	}
}

void video_drawRect(short x, short y, short w, short h, char opt){
   /* Draw a rectangle outline with top left vertex (x,y), width w
    * and height h at given color
    * Parameters:
    *      x:  x-coordinate of top-left vertex. The x-coordinate of
    *          the top-left of the screen is 0. It increases to the right.
    *      y:  y-coordinate of top-left vertex. The y-coordinate of
    *          the top-left of the screen is 0. It increases to the bottom.
    *      w:  width of the rectangle
    *      h:  height of the rectangle
    *      opt: draw option
    * Returns: Nothing
    */
	video_drawFastHLine(x, y, w, opt);
    video_drawFastHLine(x, y+h-1, w, opt);
    video_drawFastVLine(x, y, h, opt);
    video_drawFastVLine(x+w-1, y, h, opt);
}

void video_drawCircle(short x0, short y0, short r, char opt){
   /* Draw a circle outline with center (x0,y0) and radius r, with given color
    * Parameters:
    *      x0: x-coordinate of center of circle. The top-left of the screen
    *          has x-coordinate 0 and increases to the right
    *      y0: y-coordinate of center of circle. The top-left of the screen
    *          has y-coordinate 0 and increases to the bottom
    *      r:  radius of circle
    *      opt: draw option
    * Returns: Nothing
    */
     short f = 1 - r;
     short ddF_x = 1;
     short ddF_y = -2 * r;
     short x = 0;
     short y = r;

     video_pt(x0  , y0+r, opt);
     video_pt(x0  , y0-r, opt);
     video_pt(x0+r, y0  , opt);
     video_pt(x0-r, y0  , opt);

     while (x<y) {
       if (f >= 0) {
         y--;
         ddF_y += 2;
         f += ddF_y;
       }
       x++;
       ddF_x += 2;
       f += ddF_x;

       video_pt(x0 + x, y0 + y, opt);
       video_pt(x0 - x, y0 + y, opt);
       video_pt(x0 + x, y0 - y, opt);
       video_pt(x0 - x, y0 - y, opt);
       video_pt(x0 + y, y0 + x, opt);
       video_pt(x0 - y, y0 + x, opt);
       video_pt(x0 + y, y0 - x, opt);
       video_pt(x0 - y, y0 - x, opt);
     }
}

void video_drawCircleHelper(short x0, short y0, short r, unsigned char cornername, char opt){
  // Helper function for drawing circles and circular objects
    short f     = 1 - r;
    short ddF_x = 1;
    short ddF_y = -2 * r;
    short x     = 0;
    short y     = r;

    while (x<y) {
      if (f >= 0) {
        y--;
        ddF_y += 2;
        f     += ddF_y;
      }
      x++;
      ddF_x += 2;
      f     += ddF_x;
      if (cornername & 0x4) {
        video_pt(x0 + x, y0 + y, opt);
        video_pt(x0 + y, y0 + x, opt);
      }
      if (cornername & 0x2) {
        video_pt(x0 + x, y0 - y, opt);
        video_pt(x0 + y, y0 - x, opt);
      }
      if (cornername & 0x8) {
        video_pt(x0 - y, y0 + x, opt);
        video_pt(x0 - x, y0 + y, opt);
      }
      if (cornername & 0x1) {
        video_pt(x0 - y, y0 - x, opt);
        video_pt(x0 - x, y0 - y, opt);
      }
    }
}

void video_fillCircle(short x0, short y0, short r, char opt){
   /* Draw a filled circle with center (x0,y0) and radius r, with given color
    * Parameters:
    *      x0: x-coordinate of center of circle. The top-left of the screen
    *          has x-coordinate 0 and increases to the right
    *      y0: y-coordinate of center of circle. The top-left of the screen
    *          has y-coordinate 0 and increases to the bottom
    *      r:  radius of circle
    *      opt: draw option
    * Returns: Nothing
    */
     video_drawFastVLine(x0, y0-r, 2*r+1, opt);
     video_fillCircleHelper(x0, y0, r, 3, 0, opt);
}

void video_fillCircleHelper(short x0, short y0, short r, unsigned char cornername, short delta, char opt)
{
  // Helper function for drawing filled circles
    short f     = 1 - r;
    short ddF_x = 1;
    short ddF_y = -2 * r;
    short x     = 0;
    short y     = r;

    while (x<y) {
      if (f >= 0) {
        y--;
        ddF_y += 2;
        f     += ddF_y;
      }
      x++;
      ddF_x += 2;
      f     += ddF_x;

      if (cornername & 0x1) {
        video_drawFastVLine(x0+x, y0-y, 2*y+1+delta, opt);
        video_drawFastVLine(x0+y, y0-x, 2*x+1+delta, opt);
      }
      if (cornername & 0x2) {
        video_drawFastVLine(x0-x, y0-y, 2*y+1+delta, opt);
        video_drawFastVLine(x0-y, y0-x, 2*x+1+delta, opt);
      }
    }
}

void video_drawTriangle(short x0, short y0, short x1, short y1, short x2, short y2, char opt);
void video_fillTriangle(short x0, short y0, short x1, short y1, short x2, short y2, char opt);

void video_drawRoundRect(short x0, short y0, short w, short h, short r, char opt){
  // smarter version
  video_drawFastHLine(x0+r  , y0    , w-2*r, opt); // Top
  video_drawFastHLine(x0+r  , y0+h-1, w-2*r, opt); // Bottom
  video_drawFastVLine(x0    , y0+r  , h-2*r, opt); // Left
  video_drawFastVLine(x0+w-1, y0+r  , h-2*r, opt); // Right
  // draw four corners
  video_drawCircleHelper(x0+r    , y0+r    , r, 1, opt);
  video_drawCircleHelper(x0+w-r-1, y0+r    , r, 2, opt);
  video_drawCircleHelper(x0+w-r-1, y0+h-r-1, r, 4, opt);
  video_drawCircleHelper(x0+r    , y0+h-r-1, r, 8, opt);
}

void video_fillRect(short x0, short y0, short w, short h, char opt){
    short i;
    for (i = y0; i < y0+h; i++){
        video_drawFastHLine(x0, i, w, opt);
    }
}

void video_fillRoundRect(short x0, short y0, short w, short h, short r, char opt){
    // smarter version
  video_fillRect(x0+r, y0, w-2*r, h, opt);

  // draw four corners
  video_fillCircleHelper(x0+w-r-1, y0+r, r, 1, h-2*r-1, opt);
  video_fillCircleHelper(x0+r    , y0+r, r, 2, h-2*r-1, opt);
}
void video_drawBitmap(short x, short y, const unsigned char *bitmap, short w, short h, char opt);



// bezier functions
void video_drawBezier(int x0, int y0, int x1, int y1, int x2, int y2, char opt){
	fix i;
	int xa, ya, xb, yb, x, y;
	for (i = 0; i < int2fix(1); i += fixpercent){
		xa = video_drawBezierHelper(x0, x1, i);
		ya = video_drawBezierHelper(y0, y1, i);
		xb = video_drawBezierHelper(x1, x2, i);
		yb = video_drawBezierHelper(y1, y2, i);
		
		x = video_drawBezierHelper(xa, xb, i);
		y = video_drawBezierHelper(ya, yb, i);
		
		video_pt(x, y, opt);
	}
}
int video_drawBezierHelper(int n1, int n2, fix perc){
	int diff = n2 - n1;
	
	return n1 + fix2int(multfix(int2fix(diff),perc));
}